from .agent import Agent
from .agent_loop import (
    agent_loop,
    agent_loop_continue,
    agentLoop,
    agentLoopContinue,
    validate_tool_arguments,
)
from .types import AgentState

__all__ = [
    "Agent",
    "AgentState",
    "agent_loop",
    "agent_loop_continue",
    "agentLoop",
    "agentLoopContinue",
    "validate_tool_arguments",
]



from __future__ import annotations

import copy
import inspect
import time
from dataclasses import dataclass, field
from typing import (
    Any,
    Awaitable,
    Callable,
    Literal,
    Mapping,
    MutableMapping,
    Protocol,
    Required,
    Sequence,
    TypedDict,
    TypeVar,
    Union,
)

from .event_stream import AssistantMessageEventStream

ThinkingLevel = Literal["off", "minimal", "low", "medium", "high", "xhigh"]
Transport = Literal["sse", "websocket", "auto"]
StopReason = Literal["stop", "length", "toolUse", "error", "aborted"]


class ThinkingBudgets(TypedDict, total=False):
    minimal: int
    low: int
    medium: int
    high: int


class StreamCost(TypedDict):
    input: float
    output: float
    cacheRead: float
    cacheWrite: float
    total: float


class Usage(TypedDict):
    input: int
    output: int
    cacheRead: int
    cacheWrite: int
    totalTokens: int
    cost: StreamCost


class ModelCost(TypedDict):
    input: float
    output: float
    cacheRead: float
    cacheWrite: float


class Model(TypedDict, total=False):
    id: str
    name: str
    api: str
    provider: str
    baseUrl: str
    reasoning: bool
    input: list[Literal["text", "image"]]
    cost: ModelCost
    contextWindow: int
    maxTokens: int
    headers: dict[str, str]


DEFAULT_MODEL: Model = {
    "id": "gemini-2.5-flash-lite-preview-06-17",
    "name": "gemini-2.5-flash-lite-preview-06-17",
    "api": "google-generative-ai",
    "provider": "google",
    "baseUrl": "https://example.invalid",
    "reasoning": False,
    "input": ["text"],
    "cost": {"input": 0.0, "output": 0.0, "cacheRead": 0.0, "cacheWrite": 0.0},
    "contextWindow": 0,
    "maxTokens": 0,
}


class TextContent(TypedDict, total=False):
    type: Literal["text"]
    text: str
    textSignature: str


class ThinkingContent(TypedDict, total=False):
    type: Literal["thinking"]
    thinking: str
    thinkingSignature: str


class ImageContent(TypedDict):
    type: Literal["image"]
    data: str
    mimeType: str


class ToolCall(TypedDict, total=False):
    type: Literal["toolCall"]
    id: str
    name: str
    arguments: dict[str, Any]
    thoughtSignature: str


class UserMessage(TypedDict):
    role: Literal["user"]
    content: str | list[TextContent | ImageContent]
    timestamp: int


class AssistantMessage(TypedDict, total=False):
    role: Literal["assistant"]
    content: list[TextContent | ThinkingContent | ToolCall]
    api: str
    provider: str
    model: str
    usage: Usage
    stopReason: StopReason
    errorMessage: str
    timestamp: int


class ToolResultMessage(TypedDict, total=False):
    role: Literal["toolResult"]
    toolCallId: str
    toolName: str
    content: list[TextContent | ImageContent]
    details: Any
    isError: bool
    timestamp: int


Message = UserMessage | AssistantMessage | ToolResultMessage
# Custom app messages are represented as free-form mappings.
AgentMessage = Message | dict[str, Any]


class Tool(TypedDict, total=False):
    name: str
    description: str
    parameters: dict[str, Any]


class Context(TypedDict, total=False):
    systemPrompt: str
    messages: list[Message]
    tools: list[Tool]


class AssistantStartEvent(TypedDict):
    type: Literal["start"]
    partial: AssistantMessage


class AssistantTextStartEvent(TypedDict):
    type: Literal["text_start"]
    contentIndex: int
    partial: AssistantMessage


class AssistantTextDeltaEvent(TypedDict):
    type: Literal["text_delta"]
    contentIndex: int
    delta: str
    partial: AssistantMessage


class AssistantTextEndEvent(TypedDict):
    type: Literal["text_end"]
    contentIndex: int
    content: str
    partial: AssistantMessage


class AssistantThinkingStartEvent(TypedDict):
    type: Literal["thinking_start"]
    contentIndex: int
    partial: AssistantMessage


class AssistantThinkingDeltaEvent(TypedDict):
    type: Literal["thinking_delta"]
    contentIndex: int
    delta: str
    partial: AssistantMessage


class AssistantThinkingEndEvent(TypedDict):
    type: Literal["thinking_end"]
    contentIndex: int
    content: str
    partial: AssistantMessage


class AssistantToolCallStartEvent(TypedDict):
    type: Literal["toolcall_start"]
    contentIndex: int
    partial: AssistantMessage


class AssistantToolCallDeltaEvent(TypedDict):
    type: Literal["toolcall_delta"]
    contentIndex: int
    delta: str
    partial: AssistantMessage


class AssistantToolCallEndEvent(TypedDict):
    type: Literal["toolcall_end"]
    contentIndex: int
    toolCall: ToolCall
    partial: AssistantMessage


class AssistantDoneEvent(TypedDict):
    type: Literal["done"]
    reason: Literal["stop", "length", "toolUse"]
    message: AssistantMessage


class AssistantErrorEvent(TypedDict):
    type: Literal["error"]
    reason: Literal["aborted", "error"]
    error: AssistantMessage


AssistantMessageEvent = Union[
    AssistantStartEvent,
    AssistantTextStartEvent,
    AssistantTextDeltaEvent,
    AssistantTextEndEvent,
    AssistantThinkingStartEvent,
    AssistantThinkingDeltaEvent,
    AssistantThinkingEndEvent,
    AssistantToolCallStartEvent,
    AssistantToolCallDeltaEvent,
    AssistantToolCallEndEvent,
    AssistantDoneEvent,
    AssistantErrorEvent,
]


class SupportsAbort(Protocol):
    @property
    def aborted(self) -> bool: ...


class StreamResponse(Protocol):
    def __aiter__(self): ...

    async def result(self) -> AssistantMessage: ...


class AgentToolResult(TypedDict, total=False):
    content: Required[list[TextContent | ImageContent]]
    details: Any


AgentToolUpdateCallback = Callable[[AgentToolResult], None]


class AgentToolProtocol(Protocol):
    name: str
    label: str
    description: str
    parameters: Mapping[str, Any]

    async def execute(
        self,
        tool_call_id: str,
        params: Any,
        signal: SupportsAbort | None = None,
        on_update: AgentToolUpdateCallback | None = None,
    ) -> AgentToolResult: ...


AgentTool = AgentToolProtocol | Mapping[str, Any]


class AgentContext(TypedDict, total=False):
    systemPrompt: str
    messages: Required[list[AgentMessage]]
    tools: list[AgentTool]


ConvertToLlmFn = Callable[[list[AgentMessage]], Sequence[Message] | Awaitable[Sequence[Message]]]
TransformContextFn = Callable[[list[AgentMessage], SupportsAbort | None], Awaitable[list[AgentMessage]]]
GetApiKeyFn = Callable[[str], str | None | Awaitable[str | None]]
GetQueuedMessagesFn = Callable[[], Awaitable[list[AgentMessage]]]


class AgentLoopConfig(TypedDict, total=False):
    model: Required[Model]
    convertToLlm: Required[ConvertToLlmFn]
    transformContext: TransformContextFn
    getApiKey: GetApiKeyFn
    getSteeringMessages: GetQueuedMessagesFn
    getFollowUpMessages: GetQueuedMessagesFn
    reasoning: ThinkingLevel
    apiKey: str
    sessionId: str
    transport: Transport
    thinkingBudgets: ThinkingBudgets
    maxRetryDelayMs: int
    temperature: float
    maxTokens: int


class AgentStartEvent(TypedDict):
    type: Literal["agent_start"]


class AgentEndEvent(TypedDict):
    type: Literal["agent_end"]
    messages: list[AgentMessage]


class TurnStartEvent(TypedDict):
    type: Literal["turn_start"]


class TurnEndEvent(TypedDict):
    type: Literal["turn_end"]
    message: AgentMessage
    toolResults: list[ToolResultMessage]


class MessageStartEvent(TypedDict):
    type: Literal["message_start"]
    message: AgentMessage


class MessageUpdateEvent(TypedDict):
    type: Literal["message_update"]
    message: AgentMessage
    assistantMessageEvent: AssistantMessageEvent


class MessageEndEvent(TypedDict):
    type: Literal["message_end"]
    message: AgentMessage


class ToolExecutionStartEvent(TypedDict):
    type: Literal["tool_execution_start"]
    toolCallId: str
    toolName: str
    args: Any


class ToolExecutionUpdateEvent(TypedDict):
    type: Literal["tool_execution_update"]
    toolCallId: str
    toolName: str
    args: Any
    partialResult: Any


class ToolExecutionEndEvent(TypedDict):
    type: Literal["tool_execution_end"]
    toolCallId: str
    toolName: str
    result: Any
    isError: bool


AgentEvent = Union[
    AgentStartEvent,
    AgentEndEvent,
    TurnStartEvent,
    TurnEndEvent,
    MessageStartEvent,
    MessageUpdateEvent,
    MessageEndEvent,
    ToolExecutionStartEvent,
    ToolExecutionUpdateEvent,
    ToolExecutionEndEvent,
]


class SimpleStreamOptions(TypedDict, total=False):
    reasoning: ThinkingLevel
    thinkingBudgets: ThinkingBudgets
    temperature: float
    maxTokens: int
    signal: SupportsAbort
    apiKey: str
    transport: Transport
    sessionId: str
    maxRetryDelayMs: int


StreamFn = Callable[
    [Model, Context, SimpleStreamOptions | AgentLoopConfig | dict[str, Any] | None],
    AssistantMessageEventStream | StreamResponse | Awaitable[AssistantMessageEventStream | StreamResponse],
]


@dataclass(slots=True)
class AgentState:
    system_prompt: str = ""
    model: Model = field(default_factory=lambda: copy.deepcopy(DEFAULT_MODEL))
    thinking_level: ThinkingLevel = "off"
    tools: list[AgentTool] = field(default_factory=list)
    messages: list[AgentMessage] = field(default_factory=list)
    is_streaming: bool = False
    stream_message: AgentMessage | None = None
    pending_tool_calls: set[str] = field(default_factory=set)
    error: str | None = None

    @property
    def systemPrompt(self) -> str:
        return self.system_prompt

    @systemPrompt.setter
    def systemPrompt(self, value: str) -> None:
        self.system_prompt = value

    @property
    def thinkingLevel(self) -> ThinkingLevel:
        return self.thinking_level

    @thinkingLevel.setter
    def thinkingLevel(self, value: ThinkingLevel) -> None:
        self.thinking_level = value

    @property
    def isStreaming(self) -> bool:
        return self.is_streaming

    @isStreaming.setter
    def isStreaming(self, value: bool) -> None:
        self.is_streaming = value

    @property
    def streamMessage(self) -> AgentMessage | None:
        return self.stream_message

    @streamMessage.setter
    def streamMessage(self, value: AgentMessage | None) -> None:
        self.stream_message = value

    @property
    def pendingToolCalls(self) -> set[str]:
        return self.pending_tool_calls

    @pendingToolCalls.setter
    def pendingToolCalls(self, value: set[str]) -> None:
        self.pending_tool_calls = value


class AgentOptions(TypedDict, total=False):
    initialState: Mapping[str, Any] | AgentState
    convertToLlm: ConvertToLlmFn
    transformContext: TransformContextFn
    steeringMode: Literal["all", "one-at-a-time"]
    followUpMode: Literal["all", "one-at-a-time"]
    streamFn: StreamFn
    sessionId: str
    getApiKey: GetApiKeyFn
    thinkingBudgets: ThinkingBudgets
    transport: Transport
    maxRetryDelayMs: int


def now_ms() -> int:
    return int(time.time() * 1000)


def create_default_model() -> Model:
    return copy.deepcopy(DEFAULT_MODEL)


T = TypeVar("T")


def get_value(source: Mapping[str, Any] | object, *names: str, default: T | None = None) -> Any | T | None:
    for name in names:
        if isinstance(source, Mapping) and name in source:
            return source[name]
        if hasattr(source, name):
            return getattr(source, name)
    return default


def set_value(target: MutableMapping[str, Any] | object, name: str, value: Any) -> None:
    if isinstance(target, MutableMapping):
        target[name] = value
        return
    setattr(target, name, value)


def clone_message(message: AgentMessage) -> AgentMessage:
    return copy.deepcopy(message)


def clone_messages(messages: Sequence[AgentMessage]) -> list[AgentMessage]:
    return [clone_message(message) for message in messages]


def message_role(message: AgentMessage | Mapping[str, Any] | object) -> str | None:
    role = get_value(message, "role", default=None)
    return role if isinstance(role, str) else None


def message_content(message: AgentMessage | Mapping[str, Any] | object) -> Any:
    return get_value(message, "content", default=None)


def stop_reason(message: AgentMessage | Mapping[str, Any] | object) -> str | None:
    value = get_value(message, "stopReason", "stop_reason", default=None)
    return value if isinstance(value, str) else None


def model_provider(model: Model | Mapping[str, Any]) -> str:
    provider = get_value(model, "provider", default="")
    return provider if isinstance(provider, str) else ""


def ensure_usage(value: Mapping[str, Any] | None = None) -> Usage:
    if value is None:
        return {
            "input": 0,
            "output": 0,
            "cacheRead": 0,
            "cacheWrite": 0,
            "totalTokens": 0,
            "cost": {"input": 0.0, "output": 0.0, "cacheRead": 0.0, "cacheWrite": 0.0, "total": 0.0},
        }
    # Keep any caller-provided usage fields, but guarantee mandatory keys.
    usage = dict(value)
    usage.setdefault("input", 0)
    usage.setdefault("output", 0)
    usage.setdefault("cacheRead", 0)
    usage.setdefault("cacheWrite", 0)
    usage.setdefault("totalTokens", 0)
    cost = dict(usage.get("cost", {}))
    cost.setdefault("input", 0.0)
    cost.setdefault("output", 0.0)
    cost.setdefault("cacheRead", 0.0)
    cost.setdefault("cacheWrite", 0.0)
    cost.setdefault("total", 0.0)
    usage["cost"] = cost
    return usage  # type: ignore[return-value]


def is_awaitable(value: Any) -> bool:
    return inspect.isawaitable(value)


async def maybe_await(value: Any) -> Any:
    if inspect.isawaitable(value):
        return await value
    return value


def default_convert_to_llm(messages: list[AgentMessage]) -> list[Message]:
    filtered: list[Message] = []
    for message in messages:
        if message_role(message) in {"user", "assistant", "toolResult"}:
            filtered.append(message)  # type: ignore[arg-type]
    return filtered



from __future__ import annotations

import asyncio
import copy
import json
from dataclasses import dataclass
from typing import Any, Mapping, Sequence, cast

from .event_stream import EventStream
from .types import (
    AgentContext,
    AgentEvent,
    AgentLoopConfig,
    AgentMessage,
    AgentTool,
    AgentToolResult,
    AssistantMessage,
    Context,
    Message,
    Model,
    StreamFn,
    ToolCall,
    ToolResultMessage,
    clone_message,
    get_value,
    maybe_await,
    message_content,
    message_role,
    model_provider,
    now_ms,
    stop_reason,
)


def stream_simple(model: Model, context: Context, options: Mapping[str, Any] | None = None):
    raise NotImplementedError(
        "stream_simple provider integrations are not part of pidevkit.agent. Pass `stream_fn` explicitly."
    )


def agent_loop(
    prompts: Sequence[AgentMessage],
    context: AgentContext | Mapping[str, Any],
    config: AgentLoopConfig | Mapping[str, Any],
    signal: Any | None = None,
    stream_fn: StreamFn | None = None,
) -> EventStream[AgentEvent, list[AgentMessage]]:
    stream = _create_agent_stream()

    async def runner() -> None:
        new_messages: list[AgentMessage] = list(prompts)
        current_context = _build_context(context, append=prompts)

        stream.push({"type": "agent_start"})
        stream.push({"type": "turn_start"})
        for prompt in prompts:
            stream.push({"type": "message_start", "message": prompt})
            stream.push({"type": "message_end", "message": prompt})

        await _run_loop(current_context, new_messages, config, signal, stream, stream_fn)

    _spawn_runner(stream, runner())
    return stream


def agent_loop_continue(
    context: AgentContext | Mapping[str, Any],
    config: AgentLoopConfig | Mapping[str, Any],
    signal: Any | None = None,
    stream_fn: StreamFn | None = None,
) -> EventStream[AgentEvent, list[AgentMessage]]:
    messages = list(_get_context_messages(context))
    if not messages:
        raise ValueError("Cannot continue: no messages in context")

    if message_role(messages[-1]) == "assistant":
        raise ValueError("Cannot continue from message role: assistant")

    stream = _create_agent_stream()

    async def runner() -> None:
        new_messages: list[AgentMessage] = []
        current_context = _build_context(context)
        stream.push({"type": "agent_start"})
        stream.push({"type": "turn_start"})
        await _run_loop(current_context, new_messages, config, signal, stream, stream_fn)

    _spawn_runner(stream, runner())
    return stream


def _spawn_runner(stream: EventStream[AgentEvent, list[AgentMessage]], coro: asyncio.Future[Any] | Any) -> None:
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError as exc:
        raise RuntimeError("agent_loop requires an active asyncio event loop") from exc

    task = loop.create_task(cast("asyncio.coroutines.Coroutine[Any, Any, None]", coro))

    def done_callback(done_task: asyncio.Task[Any]) -> None:
        if done_task.cancelled():
            stream.end([])
            return
        exc = done_task.exception()
        if exc is not None:
            stream.set_exception(exc)

    task.add_done_callback(done_callback)
    setattr(stream, "_task", task)


def _create_agent_stream() -> EventStream[AgentEvent, list[AgentMessage]]:
    return EventStream(
        lambda event: cast(str, event.get("type", "")) == "agent_end",
        lambda event: cast(list[AgentMessage], event.get("messages", [])),
    )


def _get_context_messages(context: AgentContext | Mapping[str, Any]) -> Sequence[AgentMessage]:
    messages = get_value(context, "messages", default=[])
    if isinstance(messages, Sequence):
        return cast(Sequence[AgentMessage], messages)
    return []


def _build_context(
    context: AgentContext | Mapping[str, Any],
    append: Sequence[AgentMessage] | None = None,
) -> AgentContext:
    messages = list(_get_context_messages(context))
    if append:
        messages.extend(append)
    system_prompt = get_value(context, "systemPrompt", "system_prompt", default="")
    tools = get_value(context, "tools", default=None)
    built: AgentContext = {"systemPrompt": cast(str, system_prompt), "messages": messages}
    if isinstance(tools, list):
        built["tools"] = cast(list[AgentTool], tools)
    return built


async def _run_loop(
    current_context: AgentContext,
    new_messages: list[AgentMessage],
    config: AgentLoopConfig | Mapping[str, Any],
    signal: Any | None,
    stream: EventStream[AgentEvent, list[AgentMessage]],
    stream_fn: StreamFn | None = None,
) -> None:
    first_turn = True
    pending_messages = await _get_queued_messages(config, "getSteeringMessages", "get_steering_messages")

    while True:
        has_more_tool_calls = True
        steering_after_tools: list[AgentMessage] | None = None

        while has_more_tool_calls or pending_messages:
            if first_turn:
                first_turn = False
            else:
                stream.push({"type": "turn_start"})

            if pending_messages:
                for message in pending_messages:
                    stream.push({"type": "message_start", "message": message})
                    stream.push({"type": "message_end", "message": message})
                    current_context["messages"].append(message)
                    new_messages.append(message)
                pending_messages = []

            message = await _stream_assistant_response(current_context, config, signal, stream, stream_fn)
            new_messages.append(message)

            reason = stop_reason(message)
            if reason in {"error", "aborted"}:
                stream.push({"type": "turn_end", "message": message, "toolResults": []})
                stream.push({"type": "agent_end", "messages": new_messages})
                stream.end(new_messages)
                return

            tool_calls = _extract_tool_calls(message)
            has_more_tool_calls = len(tool_calls) > 0

            tool_results: list[ToolResultMessage] = []
            if has_more_tool_calls:
                execution = await _execute_tool_calls(
                    cast(list[AgentTool] | None, current_context.get("tools")),
                    message,
                    signal,
                    stream,
                    cast(Any, get_value(config, "getSteeringMessages", "get_steering_messages", default=None)),
                )
                tool_results.extend(execution.tool_results)
                steering_after_tools = execution.steering_messages
                for result in tool_results:
                    current_context["messages"].append(result)
                    new_messages.append(result)

            stream.push({"type": "turn_end", "message": message, "toolResults": tool_results})

            if steering_after_tools:
                pending_messages = steering_after_tools
                steering_after_tools = None
            else:
                pending_messages = await _get_queued_messages(config, "getSteeringMessages", "get_steering_messages")

        follow_up_messages = await _get_queued_messages(config, "getFollowUpMessages", "get_follow_up_messages")
        if follow_up_messages:
            pending_messages = follow_up_messages
            continue

        break

    stream.push({"type": "agent_end", "messages": new_messages})
    stream.end(new_messages)


async def _stream_assistant_response(
    context: AgentContext,
    config: AgentLoopConfig | Mapping[str, Any],
    signal: Any | None,
    stream: EventStream[AgentEvent, list[AgentMessage]],
    stream_fn: StreamFn | None = None,
) -> AssistantMessage:
    messages: list[AgentMessage] = list(context["messages"])

    transform_context = get_value(config, "transformContext", "transform_context", default=None)
    if callable(transform_context):
        transformed = await maybe_await(transform_context(messages, signal))
        messages = list(transformed)

    convert_to_llm = get_value(config, "convertToLlm", "convert_to_llm", default=None)
    if not callable(convert_to_llm):
        raise ValueError("AgentLoopConfig.convertToLlm is required")

    llm_messages = await maybe_await(convert_to_llm(messages))
    llm_context: Context = {
        "systemPrompt": cast(str, context.get("systemPrompt", "")),
        "messages": cast(list[Message], list(llm_messages)),
    }
    if "tools" in context and context["tools"] is not None:
        llm_context["tools"] = cast(list[Any], context["tools"])

    stream_function = stream_fn or stream_simple
    model = cast(Model, get_value(config, "model", default=None))
    if not model:
        raise ValueError("AgentLoopConfig.model is required")

    get_api_key = get_value(config, "getApiKey", "get_api_key", default=None)
    resolved_api_key = None
    if callable(get_api_key):
        resolved_api_key = await maybe_await(get_api_key(model_provider(model)))
    if not resolved_api_key:
        resolved_api_key = get_value(config, "apiKey", "api_key", default=None)

    options = _config_to_options(config)
    options["apiKey"] = resolved_api_key
    options["signal"] = signal

    response = await maybe_await(stream_function(model, llm_context, options))

    partial_message: AssistantMessage | None = None
    added_partial = False

    async for event in response:
        event_type = get_value(event, "type", default="")
        if event_type == "start":
            partial_message = cast(AssistantMessage, get_value(event, "partial", default={}))
            context["messages"].append(partial_message)
            added_partial = True
            stream.push({"type": "message_start", "message": clone_message(partial_message)})
            continue

        if event_type in {
            "text_start",
            "text_delta",
            "text_end",
            "thinking_start",
            "thinking_delta",
            "thinking_end",
            "toolcall_start",
            "toolcall_delta",
            "toolcall_end",
        }:
            if partial_message is not None:
                partial_message = cast(AssistantMessage, get_value(event, "partial", default=partial_message))
                context["messages"][-1] = partial_message
                stream.push(
                    {
                        "type": "message_update",
                        "assistantMessageEvent": event,
                        "message": clone_message(partial_message),
                    }
                )
            continue

        if event_type in {"done", "error"}:
            final_message = cast(AssistantMessage, await response.result())
            if added_partial:
                context["messages"][-1] = final_message
            else:
                context["messages"].append(final_message)
                stream.push({"type": "message_start", "message": clone_message(final_message)})
            stream.push({"type": "message_end", "message": final_message})
            return final_message

    return cast(AssistantMessage, await response.result())


@dataclass(slots=True)
class _ToolExecutionResult:
    tool_results: list[ToolResultMessage]
    steering_messages: list[AgentMessage] | None = None


async def _execute_tool_calls(
    tools: list[AgentTool] | None,
    assistant_message: AssistantMessage,
    signal: Any | None,
    stream: EventStream[AgentEvent, list[AgentMessage]],
    get_steering_messages: Any = None,
) -> _ToolExecutionResult:
    tool_calls = _extract_tool_calls(assistant_message)
    results: list[ToolResultMessage] = []
    steering_messages: list[AgentMessage] | None = None

    for index, tool_call in enumerate(tool_calls):
        tool_name = cast(str, tool_call.get("name", ""))
        tool_id = cast(str, tool_call.get("id", ""))
        tool_args = tool_call.get("arguments", {})
        tool = _find_tool(tools, tool_name)

        stream.push(
            {
                "type": "tool_execution_start",
                "toolCallId": tool_id,
                "toolName": tool_name,
                "args": tool_args,
            }
        )

        result: AgentToolResult
        is_error = False
        try:
            if tool is None:
                raise ValueError(f"Tool {tool_name} not found")

            validated_args = validate_tool_arguments(tool, tool_call)
            result = await _execute_tool(tool, tool_id, validated_args, signal, stream, tool_name, tool_args)
        except Exception as exc:
            result = {
                "content": [{"type": "text", "text": str(exc)}],
                "details": {},
            }
            is_error = True

        stream.push(
            {
                "type": "tool_execution_end",
                "toolCallId": tool_id,
                "toolName": tool_name,
                "result": result,
                "isError": is_error,
            }
        )

        tool_result: ToolResultMessage = {
            "role": "toolResult",
            "toolCallId": tool_id,
            "toolName": tool_name,
            "content": cast(list[Any], result.get("content", [])),
            "details": result.get("details", {}),
            "isError": is_error,
            "timestamp": now_ms(),
        }
        results.append(tool_result)
        stream.push({"type": "message_start", "message": tool_result})
        stream.push({"type": "message_end", "message": tool_result})

        if callable(get_steering_messages):
            steering = await maybe_await(get_steering_messages())
            steering_list = list(steering or [])
            if steering_list:
                steering_messages = steering_list
                for skipped in tool_calls[index + 1 :]:
                    results.append(_skip_tool_call(skipped, stream))
                break

    return _ToolExecutionResult(tool_results=results, steering_messages=steering_messages)


async def _execute_tool(
    tool: AgentTool,
    tool_call_id: str,
    validated_args: Any,
    signal: Any | None,
    stream: EventStream[AgentEvent, list[AgentMessage]],
    tool_name: str,
    tool_args: Any,
) -> AgentToolResult:
    execute = get_value(tool, "execute", default=None)
    if not callable(execute):
        raise ValueError(f"Tool {tool_name} missing execute()")

    def on_update(partial_result: AgentToolResult) -> None:
        stream.push(
            {
                "type": "tool_execution_update",
                "toolCallId": tool_call_id,
                "toolName": tool_name,
                "args": tool_args,
                "partialResult": partial_result,
            }
        )

    result = await maybe_await(execute(tool_call_id, validated_args, signal, on_update))
    if not isinstance(result, Mapping):
        raise TypeError(f"Tool {tool_name} returned invalid result type: {type(result)!r}")
    normalized = dict(result)
    normalized.setdefault("content", [])
    normalized.setdefault("details", {})
    return cast(AgentToolResult, normalized)


def _skip_tool_call(
    tool_call: ToolCall,
    stream: EventStream[AgentEvent, list[AgentMessage]],
) -> ToolResultMessage:
    tool_name = cast(str, tool_call.get("name", ""))
    tool_id = cast(str, tool_call.get("id", ""))
    args = tool_call.get("arguments", {})
    result: AgentToolResult = {
        "content": [{"type": "text", "text": "Skipped due to queued user message."}],
        "details": {},
    }

    stream.push({"type": "tool_execution_start", "toolCallId": tool_id, "toolName": tool_name, "args": args})
    stream.push(
        {
            "type": "tool_execution_end",
            "toolCallId": tool_id,
            "toolName": tool_name,
            "result": result,
            "isError": True,
        }
    )

    tool_result: ToolResultMessage = {
        "role": "toolResult",
        "toolCallId": tool_id,
        "toolName": tool_name,
        "content": cast(list[Any], result["content"]),
        "details": {},
        "isError": True,
        "timestamp": now_ms(),
    }
    stream.push({"type": "message_start", "message": tool_result})
    stream.push({"type": "message_end", "message": tool_result})
    return tool_result


def _extract_tool_calls(message: AgentMessage) -> list[ToolCall]:
    content = message_content(message)
    if not isinstance(content, list):
        return []
    tool_calls: list[ToolCall] = []
    for block in content:
        if isinstance(block, Mapping) and block.get("type") == "toolCall":
            tool_calls.append(cast(ToolCall, block))
    return tool_calls


def _find_tool(tools: Sequence[AgentTool] | None, name: str) -> AgentTool | None:
    if tools is None:
        return None
    for tool in tools:
        if get_value(tool, "name", default="") == name:
            return tool
    return None


async def _get_queued_messages(
    config: AgentLoopConfig | Mapping[str, Any],
    *keys: str,
) -> list[AgentMessage]:
    getter = get_value(config, *keys, default=None)
    if not callable(getter):
        return []
    messages = await maybe_await(getter())
    return list(messages or [])


def _config_to_options(config: AgentLoopConfig | Mapping[str, Any]) -> dict[str, Any]:
    if isinstance(config, Mapping):
        return dict(config)

    result: dict[str, Any] = {}
    for key in dir(config):
        if key.startswith("_"):
            continue
        value = getattr(config, key, None)
        if callable(value):
            continue
        result[key] = value
    return result


class _ValidationError(ValueError):
    pass


def validate_tool_arguments(tool: AgentTool, tool_call: ToolCall) -> Any:
    schema = get_value(tool, "parameters", default={})
    arguments = copy.deepcopy(cast(dict[str, Any], tool_call.get("arguments", {})))
    if not isinstance(schema, Mapping):
        return arguments

    try:
        return _coerce_schema_value(schema, arguments, path="root")
    except _ValidationError as exc:
        tool_name = cast(str, tool_call.get("name", get_value(tool, "name", default="tool")))
        pretty_args = json.dumps(arguments, indent=2, sort_keys=True, default=str)
        raise ValueError(
            f'Validation failed for tool "{tool_name}":\n  - {exc}\n\nReceived arguments:\n{pretty_args}'
        ) from exc


def _coerce_schema_value(schema: Mapping[str, Any], value: Any, *, path: str) -> Any:
    if "const" in schema and value != schema["const"]:
        raise _ValidationError(f"{path}: expected constant {schema['const']!r}")

    if "enum" in schema:
        enum_values = schema["enum"]
        if isinstance(enum_values, list) and value not in enum_values:
            raise _ValidationError(f"{path}: expected one of {enum_values!r}")

    for composite_key in ("oneOf", "anyOf"):
        if composite_key in schema and isinstance(schema[composite_key], list):
            errors: list[str] = []
            for option in schema[composite_key]:
                if not isinstance(option, Mapping):
                    continue
                try:
                    return _coerce_schema_value(option, value, path=path)
                except _ValidationError as err:
                    errors.append(str(err))
            if errors:
                raise _ValidationError(f"{path}: no matching {composite_key} schema ({'; '.join(errors)})")

    schema_type = schema.get("type")
    if isinstance(schema_type, list):
        errors: list[str] = []
        for option_type in schema_type:
            option_schema = dict(schema)
            option_schema["type"] = option_type
            try:
                return _coerce_schema_value(option_schema, value, path=path)
            except _ValidationError as err:
                errors.append(str(err))
        raise _ValidationError(f"{path}: {', '.join(errors)}")

    if schema_type == "object" or ("properties" in schema and schema_type is None):
        if not isinstance(value, Mapping):
            raise _ValidationError(f"{path}: expected object")

        properties = schema.get("properties", {})
        if not isinstance(properties, Mapping):
            properties = {}
        required = schema.get("required", [])
        if isinstance(required, list):
            for field_name in required:
                if field_name not in value:
                    raise _ValidationError(f"{path}: missing required field {field_name!r}")

        additional_properties = schema.get("additionalProperties", True)
        result: dict[str, Any] = {}
        for key, raw_item in value.items():
            if key in properties and isinstance(properties[key], Mapping):
                result[key] = _coerce_schema_value(cast(Mapping[str, Any], properties[key]), raw_item, path=f"{path}.{key}")
                continue

            if additional_properties is False:
                raise _ValidationError(f"{path}: unexpected field {key!r}")
            if isinstance(additional_properties, Mapping):
                result[key] = _coerce_schema_value(additional_properties, raw_item, path=f"{path}.{key}")
            else:
                result[key] = raw_item

        return result

    if schema_type == "array":
        if not isinstance(value, list):
            raise _ValidationError(f"{path}: expected array")
        items_schema = schema.get("items")
        coerced_items = value
        if isinstance(items_schema, Mapping):
            coerced_items = [
                _coerce_schema_value(items_schema, item, path=f"{path}[{index}]") for index, item in enumerate(value)
            ]

        min_items = schema.get("minItems")
        max_items = schema.get("maxItems")
        if isinstance(min_items, int) and len(coerced_items) < min_items:
            raise _ValidationError(f"{path}: expected at least {min_items} items")
        if isinstance(max_items, int) and len(coerced_items) > max_items:
            raise _ValidationError(f"{path}: expected at most {max_items} items")
        return coerced_items

    if schema_type == "string":
        if value is None:
            raise _ValidationError(f"{path}: expected string")
        coerced = value if isinstance(value, str) else str(value)
        min_length = schema.get("minLength")
        max_length = schema.get("maxLength")
        if isinstance(min_length, int) and len(coerced) < min_length:
            raise _ValidationError(f"{path}: expected minimum length {min_length}")
        if isinstance(max_length, int) and len(coerced) > max_length:
            raise _ValidationError(f"{path}: expected maximum length {max_length}")
        return coerced

    if schema_type == "integer":
        coerced = _coerce_number(value, integer=True, path=path)
        _validate_numeric_bounds(schema, coerced, path=path)
        return coerced

    if schema_type == "number":
        coerced = _coerce_number(value, integer=False, path=path)
        _validate_numeric_bounds(schema, coerced, path=path)
        return coerced

    if schema_type == "boolean":
        if isinstance(value, bool):
            return value
        if isinstance(value, int) and value in {0, 1}:
            return bool(value)
        if isinstance(value, str):
            lowered = value.strip().lower()
            if lowered in {"true", "1"}:
                return True
            if lowered in {"false", "0"}:
                return False
        raise _ValidationError(f"{path}: expected boolean")

    if schema_type == "null":
        if value is None:
            return None
        raise _ValidationError(f"{path}: expected null")

    return value


def _coerce_number(value: Any, *, integer: bool, path: str) -> int | float:
    if isinstance(value, bool):
        raise _ValidationError(f"{path}: expected {'integer' if integer else 'number'}")

    if integer:
        if isinstance(value, int):
            return value
        if isinstance(value, float) and value.is_integer():
            return int(value)
        if isinstance(value, str):
            stripped = value.strip()
            try:
                parsed = float(stripped)
            except ValueError as exc:
                raise _ValidationError(f"{path}: expected integer") from exc
            if not parsed.is_integer():
                raise _ValidationError(f"{path}: expected integer")
            return int(parsed)
        raise _ValidationError(f"{path}: expected integer")

    if isinstance(value, (int, float)) and not isinstance(value, bool):
        return float(value)
    if isinstance(value, str):
        stripped = value.strip()
        try:
            return float(stripped)
        except ValueError as exc:
            raise _ValidationError(f"{path}: expected number") from exc
    raise _ValidationError(f"{path}: expected number")


def _validate_numeric_bounds(schema: Mapping[str, Any], value: int | float, *, path: str) -> None:
    minimum = schema.get("minimum")
    maximum = schema.get("maximum")
    exclusive_minimum = schema.get("exclusiveMinimum")
    exclusive_maximum = schema.get("exclusiveMaximum")
    if isinstance(minimum, (int, float)) and value < minimum:
        raise _ValidationError(f"{path}: expected >= {minimum}")
    if isinstance(maximum, (int, float)) and value > maximum:
        raise _ValidationError(f"{path}: expected <= {maximum}")
    if isinstance(exclusive_minimum, (int, float)) and value <= exclusive_minimum:
        raise _ValidationError(f"{path}: expected > {exclusive_minimum}")
    if isinstance(exclusive_maximum, (int, float)) and value >= exclusive_maximum:
        raise _ValidationError(f"{path}: expected < {exclusive_maximum}")


def agentLoop(
    prompts: Sequence[AgentMessage],
    context: AgentContext | Mapping[str, Any],
    config: AgentLoopConfig | Mapping[str, Any],
    signal: Any | None = None,
    stream_fn: StreamFn | None = None,
) -> EventStream[AgentEvent, list[AgentMessage]]:
    return agent_loop(prompts, context, config, signal=signal, stream_fn=stream_fn)


def agentLoopContinue(
    context: AgentContext | Mapping[str, Any],
    config: AgentLoopConfig | Mapping[str, Any],
    signal: Any | None = None,
    stream_fn: StreamFn | None = None,
) -> EventStream[AgentEvent, list[AgentMessage]]:
    return agent_loop_continue(context, config, signal=signal, stream_fn=stream_fn)



from __future__ import annotations

import asyncio
import copy
from dataclasses import dataclass
from typing import Any, Callable, Mapping

from .agent_loop import agent_loop, agent_loop_continue
from .types import (
    AgentMessage,
    AgentOptions,
    AgentState,
    DEFAULT_MODEL,
    Message,
    Model,
    ThinkingBudgets,
    ThinkingLevel,
    Transport,
    clone_message,
    clone_messages,
    get_value,
    message_content,
    message_role,
    now_ms,
)


@dataclass(slots=True)
class AbortSignal:
    aborted: bool = False


class AbortController:
    def __init__(self) -> None:
        self.signal = AbortSignal()

    def abort(self) -> None:
        self.signal.aborted = True


class Agent:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        options: dict[str, Any] = {}
        if args:
            if len(args) > 1:
                raise TypeError("Agent accepts at most one positional options argument")
            if not isinstance(args[0], Mapping):
                raise TypeError("Agent positional argument must be a mapping")
            options.update(dict(args[0]))
        options.update(kwargs)

        self._listeners: list[Callable[[dict[str, Any]], None]] = []
        self._steering_queue: list[AgentMessage] = []
        self._follow_up_queue: list[AgentMessage] = []
        self._abort_controller: AbortController | None = None

        self._stream_fn = get_value(options, "stream_fn", "streamFn", default=None)
        self._convert_to_llm = get_value(options, "convert_to_llm", "convertToLlm", default=self._default_convert_to_llm)
        self._transform_context = get_value(options, "transform_context", "transformContext", default=None)
        self._get_api_key = get_value(options, "get_api_key", "getApiKey", default=None)

        self._steering_mode = get_value(options, "steering_mode", "steeringMode", default="one-at-a-time")
        self._follow_up_mode = get_value(options, "follow_up_mode", "followUpMode", default="one-at-a-time")

        self.thinking_budgets: ThinkingBudgets = get_value(
            options,
            "thinking_budgets",
            "thinkingBudgets",
            default={"minimal": 0, "low": 0, "medium": 0, "high": 0},
        )
        self.transport: Transport = get_value(options, "transport", default="sse")
        self.max_retry_delay_ms: int = int(get_value(options, "max_retry_delay_ms", "maxRetryDelayMs", default=1000) or 1000)

        self._session_id: str | None = get_value(options, "session_id", "sessionId", default=None)

        initial_state = get_value(options, "initial_state", "initialState", default=None)
        self._state = self._build_initial_state(initial_state)

    @property
    def state(self) -> AgentState:
        return self._state

    def subscribe(self, listener: Callable[[dict[str, Any]], None]) -> Callable[[], None]:
        self._listeners.append(listener)

        def unsubscribe() -> None:
            if listener in self._listeners:
                self._listeners.remove(listener)

        return unsubscribe

    def _emit(self, event: dict[str, Any]) -> None:
        for listener in list(self._listeners):
            try:
                listener(event)
            except Exception:
                continue

    def set_system_prompt(self, system_prompt: str) -> None:
        self._state.system_prompt = system_prompt

    def setSystemPrompt(self, system_prompt: str) -> None:
        self.set_system_prompt(system_prompt)

    def set_model(self, model: Model) -> None:
        self._state.model = copy.deepcopy(model)

    def setModel(self, model: Model) -> None:
        self.set_model(model)

    def set_thinking_level(self, level: ThinkingLevel) -> None:
        self._state.thinking_level = level

    def setThinkingLevel(self, level: ThinkingLevel) -> None:
        self.set_thinking_level(level)

    def set_tools(self, tools: list[Any]) -> None:
        self._state.tools = list(tools)

    def setTools(self, tools: list[Any]) -> None:
        self.set_tools(tools)

    def replace_messages(self, messages: list[AgentMessage]) -> None:
        self._state.messages = clone_messages(messages)

    def replaceMessages(self, messages: list[AgentMessage]) -> None:
        self.replace_messages(messages)

    def append_message(self, message: AgentMessage) -> None:
        self._state.messages.append(clone_message(message))

    def appendMessage(self, message: AgentMessage) -> None:
        self.append_message(message)

    def clear_messages(self) -> None:
        self._state.messages = []

    def clearMessages(self) -> None:
        self.clear_messages()

    def steer(self, message: AgentMessage) -> None:
        self._steering_queue.append(clone_message(message))

    def follow_up(self, message: AgentMessage) -> None:
        self._follow_up_queue.append(clone_message(message))

    def followUp(self, message: AgentMessage) -> None:
        self.follow_up(message)

    def abort(self) -> None:
        if self._abort_controller is not None:
            self._abort_controller.abort()

    @property
    def session_id(self) -> str | None:
        return self._session_id

    @session_id.setter
    def session_id(self, value: str | None) -> None:
        self._session_id = value

    @property
    def sessionId(self) -> str | None:
        return self._session_id

    @sessionId.setter
    def sessionId(self, value: str | None) -> None:
        self._session_id = value

    async def prompt(self, input_value: str | AgentMessage, images: list[dict[str, Any]] | None = None) -> list[AgentMessage]:
        self._ensure_not_streaming()
        user_message = self._make_user_message(input_value, images=images)
        return await self._run_loop(prompts=[user_message], is_continue=False)

    async def continue_(self) -> list[AgentMessage]:
        self._ensure_not_streaming()

        queued_steering = self._dequeue(self._steering_queue, self._steering_mode)
        if queued_steering:
            return await self._run_loop(prompts=queued_steering, is_continue=False)

        queued_follow_up = self._dequeue(self._follow_up_queue, self._follow_up_mode)
        if queued_follow_up:
            return await self._run_loop(prompts=queued_follow_up, is_continue=False)

        return await self._run_loop(prompts=[], is_continue=True)

    async def resume(self) -> list[AgentMessage]:
        return await self.continue_()

    async def run_continue(self) -> list[AgentMessage]:
        return await self.continue_()

    async def wait_for_idle(self) -> None:
        while self._state.is_streaming:
            await asyncio.sleep(0.005)

    async def _run_loop(self, prompts: list[AgentMessage], *, is_continue: bool) -> list[AgentMessage]:
        self._state.is_streaming = True
        self._state.error = None
        self._state.stream_message = None
        self._state.pending_tool_calls.clear()

        self._abort_controller = AbortController()
        skip_initial_steering_poll = len(prompts) > 0
        steering_polled = False

        async def get_steering_messages() -> list[AgentMessage]:
            nonlocal steering_polled
            if skip_initial_steering_poll and not steering_polled:
                steering_polled = True
                return []
            steering_polled = True
            return self._dequeue(self._steering_queue, self._steering_mode)

        async def get_follow_up_messages() -> list[AgentMessage]:
            return self._dequeue(self._follow_up_queue, self._follow_up_mode)

        context: dict[str, Any] = {
            "systemPrompt": self._state.system_prompt,
            "messages": clone_messages(self._state.messages),
            "tools": list(self._state.tools),
        }
        config = self._build_loop_config(
            get_steering_messages=get_steering_messages,
            get_follow_up_messages=get_follow_up_messages,
        )

        try:
            if is_continue:
                stream = agent_loop_continue(
                    context,
                    config,
                    signal=self._abort_controller.signal,
                    stream_fn=self._stream_fn,
                )
            else:
                stream = agent_loop(
                    prompts,
                    context,
                    config,
                    signal=self._abort_controller.signal,
                    stream_fn=self._stream_fn,
                )

            async for event in stream:
                self._handle_stream_event(event)
                self._emit(event)

            new_messages = await stream.result()
            if not isinstance(new_messages, list):
                new_messages = list(new_messages)

            self._state.messages.extend(clone_messages(new_messages))
            return new_messages
        except Exception as exc:
            self._state.error = str(exc)
            raise
        finally:
            self._state.is_streaming = False
            self._state.stream_message = None
            self._state.pending_tool_calls.clear()
            self._abort_controller = None

    def _handle_stream_event(self, event: Mapping[str, Any]) -> None:
        event_type = str(get_value(event, "type", default=""))

        if event_type in {"message_start", "message_update"}:
            message = get_value(event, "message", default=None)
            if message_role(message) == "assistant":
                self._state.stream_message = clone_message(message)
                self._sync_pending_tool_calls(message)

        if event_type == "message_end":
            message = get_value(event, "message", default=None)
            if message_role(message) == "assistant":
                self._state.stream_message = None

        if event_type == "tool_execution_start":
            tool_call_id = get_value(event, "toolCallId", "tool_call_id", default=None)
            if isinstance(tool_call_id, str):
                self._state.pending_tool_calls.add(tool_call_id)

        if event_type == "tool_execution_end":
            tool_call_id = get_value(event, "toolCallId", "tool_call_id", default=None)
            if isinstance(tool_call_id, str):
                self._state.pending_tool_calls.discard(tool_call_id)

    def _sync_pending_tool_calls(self, message: AgentMessage) -> None:
        content = message_content(message)
        if not isinstance(content, list):
            return

        for block in content:
            if isinstance(block, Mapping) and block.get("type") == "toolCall":
                tool_call_id = block.get("id")
                if isinstance(tool_call_id, str):
                    self._state.pending_tool_calls.add(tool_call_id)

    def _build_loop_config(
        self,
        *,
        get_steering_messages: Callable[[], Any] | None = None,
        get_follow_up_messages: Callable[[], Any] | None = None,
    ) -> dict[str, Any]:
        config: dict[str, Any] = {
            "model": self._state.model,
            "convertToLlm": self._convert_to_llm,
            "getSteeringMessages": get_steering_messages or self._get_steering_messages,
            "getFollowUpMessages": get_follow_up_messages or self._get_follow_up_messages,
            "reasoning": self._state.thinking_level,
            "transport": self.transport,
            "thinkingBudgets": self.thinking_budgets,
            "maxRetryDelayMs": self.max_retry_delay_ms,
        }

        if callable(self._transform_context):
            config["transformContext"] = self._transform_context

        if callable(self._get_api_key):
            config["getApiKey"] = self._get_api_key

        if self._session_id is not None:
            config["sessionId"] = self._session_id
            config["session_id"] = self._session_id

        return config

    async def _get_steering_messages(self) -> list[AgentMessage]:
        return self._dequeue(self._steering_queue, self._steering_mode)

    async def _get_follow_up_messages(self) -> list[AgentMessage]:
        return self._dequeue(self._follow_up_queue, self._follow_up_mode)

    @staticmethod
    def _dequeue(queue: list[AgentMessage], mode: str) -> list[AgentMessage]:
        if not queue:
            return []

        if mode == "all":
            values = queue[:]
            queue.clear()
            return values

        return [queue.pop(0)]

    @staticmethod
    def _default_convert_to_llm(messages: list[AgentMessage]) -> list[Message]:
        converted: list[Message] = []
        for message in messages:
            role = message_role(message)
            if role in {"user", "assistant", "toolResult"}:
                converted.append(copy.deepcopy(message))
        return converted

    def _build_initial_state(self, initial_state: Mapping[str, Any] | AgentState | None) -> AgentState:
        if isinstance(initial_state, AgentState):
            return copy.deepcopy(initial_state)

        state = AgentState()
        if initial_state is None:
            return state

        state.system_prompt = str(get_value(initial_state, "system_prompt", "systemPrompt", default="") or "")

        model_value = get_value(initial_state, "model", default=None)
        if isinstance(model_value, Mapping):
            state.model = copy.deepcopy(dict(model_value))
        else:
            state.model = copy.deepcopy(DEFAULT_MODEL)

        state.thinking_level = get_value(initial_state, "thinking_level", "thinkingLevel", default="off")

        tools_value = get_value(initial_state, "tools", default=[])
        state.tools = list(tools_value) if isinstance(tools_value, list) else []

        messages_value = get_value(initial_state, "messages", default=[])
        state.messages = clone_messages(messages_value if isinstance(messages_value, list) else [])

        state.is_streaming = bool(get_value(initial_state, "is_streaming", "isStreaming", default=False))
        state.stream_message = clone_message(get_value(initial_state, "stream_message", "streamMessage", default=None))
        pending_value = get_value(initial_state, "pending_tool_calls", "pendingToolCalls", default=set())
        if isinstance(pending_value, set):
            state.pending_tool_calls = set(str(item) for item in pending_value)
        elif isinstance(pending_value, list):
            state.pending_tool_calls = set(str(item) for item in pending_value)
        else:
            state.pending_tool_calls = set()
        state.error = get_value(initial_state, "error", default=None)

        return state

    def _ensure_not_streaming(self) -> None:
        if self._state.is_streaming:
            raise RuntimeError("Agent is already processing a request")

    @staticmethod
    def _make_user_message(input_value: str | AgentMessage, images: list[dict[str, Any]] | None = None) -> AgentMessage:
        if isinstance(input_value, Mapping) and message_role(input_value):
            message = clone_message(input_value)
            if get_value(message, "timestamp", default=None) is None:
                message["timestamp"] = now_ms()
            return message

        if isinstance(input_value, str):
            if images:
                content: list[dict[str, Any]] = [{"type": "text", "text": input_value}, *images]
            else:
                content = input_value
            return {"role": "user", "content": content, "timestamp": now_ms()}

        raise TypeError("prompt() expects a string or user message mapping")


def _resolve_options_map(options: AgentOptions | Mapping[str, Any]) -> Mapping[str, Any]:
    if isinstance(options, Mapping):
        return options
    return dict(options)



from __future__ import annotations

import asyncio
from collections import deque
from dataclasses import dataclass
from typing import Any, AsyncIterator, Callable, Deque, Generic, TypeVar, cast

TEvent = TypeVar("TEvent")
TResult = TypeVar("TResult")


@dataclass(frozen=True, slots=True)
class _Done:
    pass


_DONE = _Done()


class EventStream(Generic[TEvent, TResult]):
    """Generic push-based async event stream with a separately awaitable final result."""

    def __init__(
        self,
        is_complete: Callable[[TEvent], bool],
        extract_result: Callable[[TEvent], TResult],
    ) -> None:
        self._is_complete = is_complete
        self._extract_result = extract_result
        self._queue: Deque[TEvent] = deque()
        self._waiting: Deque[asyncio.Future[TEvent | _Done]] = deque()
        self._done = False
        self._final_result: asyncio.Future[TResult] = self._new_future()

    def _new_future(self) -> asyncio.Future[Any]:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.get_event_loop()
        return loop.create_future()

    def push(self, event: TEvent) -> None:
        if self._done:
            return

        if self._is_complete(event):
            self._done = True
            if not self._final_result.done():
                self._final_result.set_result(self._extract_result(event))

        if self._waiting:
            waiter = self._waiting.popleft()
            if not waiter.done():
                waiter.set_result(event)
            return

        self._queue.append(event)

    def end(self, result: TResult | None = None) -> None:
        self._done = True
        if result is not None and not self._final_result.done():
            self._final_result.set_result(result)
        self._finish_waiters()

    def set_exception(self, exc: BaseException) -> None:
        self._done = True
        if not self._final_result.done():
            self._final_result.set_exception(exc)
        self._finish_waiters()

    def _finish_waiters(self) -> None:
        while self._waiting:
            waiter = self._waiting.popleft()
            if not waiter.done():
                waiter.set_result(_DONE)

    async def result(self) -> TResult:
        return await self._final_result

    def __aiter__(self) -> AsyncIterator[TEvent]:
        return self._iterate()

    async def _iterate(self) -> AsyncIterator[TEvent]:
        while True:
            if self._queue:
                yield self._queue.popleft()
                continue

            if self._done:
                return

            waiter: asyncio.Future[TEvent | _Done] = cast(asyncio.Future[TEvent | _Done], self._new_future())
            self._waiting.append(waiter)
            item = await waiter
            if isinstance(item, _Done):
                return
            yield item


class AssistantMessageEventStream(EventStream[dict[str, Any], dict[str, Any]]):
    """Specialized stream whose final result is extracted from done/error events."""

    def __init__(self) -> None:
        super().__init__(
            lambda event: event.get("type") in {"done", "error"},
            lambda event: event.get("message") if event.get("type") == "done" else event.get("error"),
        )


def create_assistant_message_event_stream() -> AssistantMessageEventStream:
    return AssistantMessageEventStream()
